"""
First tested edition on May 04, 2021.
@author Jeppe Sinkbæk Thomsen, Master's student in astronomy at Aarhus University.
Supervisor: Assistant Professor Karsten Frank Brogaard.

Purpose of this file is broadening function calculation using a Singular Value Decomposition of a template spectrum.
Code is primarily adapted from http://www.astro.utoronto.ca/~rucinski/SVDcookbook.html.
The primary class to create objects from for broadening function calculation is BroadeningFunction. The rest are
convenience classes for that one.
Some of the import statements lacking below are imported from rotational_broadening_function_fitting.py.
"""

import scipy.linalg as lg
import warnings
from sb2sep.rotational_broadening_function_fitting import *
from sb2sep.storage_classes import FitParameters
import matplotlib.pyplot as plt
from copy import copy


class DesignMatrix:
    def __init__(self, template_spectrum: np.ndarray, span: int):
        """
        Creates a Design Matrix (DesignMatrix.mat) of a template spectrum for the SVD broadening function method.
        :param template_spectrum:   np.ndarray, flux of the template spectrum that design matrix should be made on.
        :param span:                int, span or width of the broadening function. Should be odd.

        :var self.vals:     same as template_spectrum
        :var self.span:     same as span
        :var self.n:        int, size of template_spectrum
        :var self.m:        same as self.span
        :var self.mat:      np.ndarray, the created design matrix. Shape (m, n-m)
        """
        self.vals = template_spectrum
        self.span = span
        self.n = self.vals.size
        self.m = self.span
        self.mat = self.map()

    def map(self):
        """
        Map stored spectrum to a design matrix.
        :return mat: np.ndarray, the created design matrix. Shape (m, n-m)
        """
        # Matrix is shape (m, n-m)
        n, m = self.n, self.m

        if np.mod(m, 2) != 1.0:
            raise ValueError('Design Matrix span must be odd.')
        if np.mod(n, 2) != 0.0:
            raise ValueError('Number of values must be even.')

        mat = np.zeros(shape=(m, n-m+1))
        for i in range(0, m):
            mat[i, :] = self.vals[i:i+n-m+1]
        return mat.T


class SingularValueDecomposition:
    def __init__(self, template_spectrum: np.ndarray, span: int):
        """
        Creates a Singular Value Decomposition of a template spectrum DesignMatrix for the SVD broadening function.

        :param template_spectrum: np.ndarray, inverted flux of the template spectrum equi-spaced in velocity space
        :param span:              int, span or width (number of elements) of the broadening function design matrix.

        :var self.design_matrix: The created DesignMatrix of the template spectrum.
        :var self.u:             np.ndarray, the unitary matrix having left singular vectors as columns.
                                 Shape (span, span)
        :var self.w:             np.ndarray, vector of the singular values, in non-increasing order.
                                 Shape (template_spectrum.size, )
        :var self.vH:            np.ndarray, the unitary matrix having right singular vectors as rows.
                                 Shape (span, template_spectrum.size)
        """
        self.design_matrix = DesignMatrix(template_spectrum, span)
        self.u, self.w, self.vH = lg.svd(self.design_matrix.mat, compute_uv=True, full_matrices=False)

    def plot_w(self):
        plt.figure()
        plt.plot(self.w)
        plt.yscale('log')
        plt.show()


class BroadeningFunction:
    def __init__(
            self,
            program_spectrum: np.ndarray,
            template_spectrum: np.ndarray,
            velocity_span: float,
            dv: float,
            span=None, copy=False, plot_w=False
    ):
        """
        Creates a broadening function object storing all the necessary variables for it.
        Using BroadeningFunction.solve(), the broadening function is found by solving the Singular Value
        Decomposition of a template spectrum.

        :param program_spectrum:  np.ndarray, inverted flux of the program spectrum equi-spaced in velocity space
        :param template_spectrum: np.ndarray, inverted flux of the template spectrum equi-spaced in velocity space
        :param velocity_span:     float, span or width (number of elements) of the broadening function design matrix.
                                    in velocity units
        :param dv:                float, the velocity spectrum resolution in km/s
        :param span:              int, span of design matrix in index units. Overrides velocity_span if not None.
        :param copy:              bool, controls if initial calculations should not be done (in case the result is to be
                                  copied from another object)

        :var self.spectrum:  same as program_spectrum
        :var self.svd:       a created SingularValueDecomposition of the template spectrum.
        :var self.bf:        None (before self.solve() is run). After, np.ndarray, the calculated broadening function.
        :var self.bf_smooth: None (before self.smooth() is run). After, np.ndarray, smoothed broadening function.
        :var self.velocity:  np.ndarray, velocity values of the broadening function in km/s.
        """
        if span is None:
            span = int(velocity_span/dv)
            self.velocity_span = velocity_span
        else:
            self.velocity_span = span*dv
        if np.mod(span, 2) != 1.0:
            warnings.warn('Design Matrix span must be odd. Shortening by 1.')
            span -= 1
        if np.mod(template_spectrum.size, 2) != 0.0:
            warnings.warn('template_spectrum length must be even. Shortening by 1.')
            template_spectrum = template_spectrum[:-1]
            program_spectrum = program_spectrum[:-1]
        if template_spectrum.size != program_spectrum.size:
            raise ValueError(f'template_spectrum.size does not match program_spectrum.size. Size '
                             f'{template_spectrum.size} vs Size {program_spectrum.size}')
        self.__spectrum = program_spectrum
        self.template_spectrum = template_spectrum
        self.span = span
        self.dv = dv
        if ~copy:
            self.svd = SingularValueDecomposition(template_spectrum, span)
        else:
            self.svd = None
        self.bf = None
        self.bf_smooth = None
        self.smooth_sigma = 5.0
        if ~copy:
            self.velocity = -np.arange(-int(span/2), int(span/2+1))*dv
        else:
            self.velocity = None
        self.fit = None
        self.model_values = None

        if plot_w and ~copy:
            self.svd.plot_w()

    @property
    def spectrum(self):
        return self.__spectrum

    @spectrum.setter
    def spectrum(self, new_value: np.ndarray):
        if np.mod(new_value.size, 2) != 0.0:
            warnings.warn('new_value for BroadeningFunction.spectrum size must be even. Shortening by 1.')
            new_value = new_value[:-1]
        self.__spectrum = new_value

    def __copy__(self):
        new_copy = type(self)(self.spectrum, self.template_spectrum, self.velocity_span, self.dv, self.span, copy=True)
        new_copy.svd = copy(self.svd)
        new_copy.bf = self.bf
        new_copy.bf_smooth = self.bf_smooth
        new_copy.smooth_sigma = self.smooth_sigma
        new_copy.velocity = self.velocity
        new_copy.fit = self.fit
        new_copy.model_values = self.model_values
        return new_copy

    @staticmethod
    def truncate(spectrum: np.ndarray, design_matrix: DesignMatrix):
        """
        Truncates a program spectrum, which is essential before solving the linear equations.
        :param spectrum:        np.ndarray, the input spectrum
        :param design_matrix:   DesignMatrix, the design matrix which is to be used
        """
        m = design_matrix.m
        truncated_spectrum = spectrum[int((m-1)/2):-int((m+1)/2)+1]
        if truncated_spectrum.size != spectrum.size - m + 1:
            print(truncated_spectrum.size)
            raise ValueError('bf truncate spectrum: Wrong length')
        return truncated_spectrum      # Found error in some cases with length. Should be fixed now

    def solve(self):
        """
        Solves the system involving the Singular Value Decomposition of the template spectrum, and the program spectrum,
        to provide the broadening function of the program spectrum.
        :return: np.ndarray, broadening function of the program spectrum
        """
        spectrum_truncated = self.truncate(self.spectrum, self.svd.design_matrix)
        u, w, vH = self.svd.u, self.svd.w, self.svd.vH

        # Safety check
        limit_w = 0.0
        w_inverse = 1.0/w
        limit_mask = (w < limit_w)
        w_inverse[limit_mask] = 0.0
        diag_mat_w_inverse = np.diag(w_inverse)

        # Matrix A: transpose(vH) diag(w_inverse) transpose(u)
        A = np.dot(vH.T, np.dot(diag_mat_w_inverse, u.T))

        # Solve linear equation to calculate broadening function
        broadening_function = np.dot(A, spectrum_truncated.T)
        self.bf = np.ravel(broadening_function)
        return self.bf

    def smooth(self):
        """
        Smoothes a calculated broadening function by convolving with a gaussian function.
        """
        if self.bf is None:
            raise TypeError('self.bf is None. self.solve() must be run prior to smoothing the broadening function.')
        gaussian = np.exp(-0.5 * (self.velocity/self.smooth_sigma)**2)
        gaussian /= np.sum(gaussian)
        self.bf_smooth = fftconvolve(self.bf, gaussian, mode='same')
        return self.bf_smooth

    def fit_rotational_profile(
            self,
            fitparams: FitParameters,
            fitting_routine=fitting_routine_rotational_broadening_profile
    ):
        """
        Fits the broadening function with a rotational broadening profile by calling a fitting routine provided.
        The routine must include all essential parts of the fitting procedure.
        :param ifitparams:          an object holding the initial fit parameters:
               vsini:               float, guess for the v sin(i) parameter
               limbd_coef:          float, a calculated limb darkening coefficient for the star.
                                         Default routine will not fit this parameter.
               velocity_fit_width:  float, how far out the fitting routine should include data-points for the fit.
               spectral_resolution: float/int, the resolution of the spectrograph used for the program spectrum.
        :param fitting_routine:     function, the fitting routine used. Default routine fits 1 rotational broadening
                                    profile to the data following the model provided by Kaluzny 2006: "Eclipsing
                                    Binaries in the Open Cluster NGC 2243 II. Absolute Properties of NV CMa".
        :return (fit, model):       fit: lmfit.MinimizerResult of the performed fit.
                                    model: np.ndarray, model values of the broadening function according to the fit.
        """
        if self.bf_smooth is None:
            raise TypeError('self.bf_smooth. self.smooth() must be run prior to fitting')

        self.fit, self.model_values = fitting_routine(
            self.velocity, self.bf_smooth, fitparams, self.smooth_sigma, self.dv
        )
        return self.fit, self.model_values






import numpy as np
import scipy.interpolate as interp


def estimate_photometric_system(wavelength_range: np.ndarray or tuple):
    """
    Estimates which photometric system is appropriate to look up linear limb-darkening coefficients for.
    Wavelengths in Ångstrom units. Assumes passbands symmetric and simple, and follows peak wvl and fwhm from
    https://en.wikipedia.org/wiki/Photometric_system and
    https://en.wikipedia.org/wiki/Str%C3%B6mgren_photometric_system
    :param wavelength_range: np.ndarray([wavelength_a, wavelength_b]) limits on the wavelength range of spectrum.
    :return: string, pass-band indicator
    """
    passband_names = np.array(['u', 'v', 'b', 'y', 'U', 'B', 'V', 'R', 'I', 'J', 'H', 'K'])
    peak_wavelength = np.array([3500, 4110, 4670, 5470, 3650, 4450, 5510, 6580, 8060, 12200, 16300, 21900])
    fwhm = np.array([300, 190, 180, 230, 660, 940, 880, 1380, 1490, 2130, 3070, 3900])

    bands_wla = peak_wavelength - fwhm
    bands_wlb = peak_wavelength + fwhm

    diff_wla = np.abs(wavelength_range[0] - bands_wla)
    diff_wlb = np.abs(wavelength_range[1] - bands_wlb)

    best_pasband_idx = np.argmin(diff_wla + diff_wlb)
    return passband_names[best_pasband_idx]


def read_linear_limbd_param(logg_range=np.array([0, 7]), Trange=np.array([2000, 9000]), MH_range=-0.5, mTurb_range=2.0,
                            loc='../Data/tables/atlasco.dat', band='V'):
    if 'atlasco.dat' in loc:
        if band=='u': band_col=5
        elif band=='v': band_col=6
        elif band=='b': band_col=7
        elif band=='y': band_col=8
        elif band=='U': band_col=9
        elif band=='B': band_col=10
        elif band=='V': band_col=11
        elif band=='R': band_col=12
        elif band=='I': band_col=13
        elif band=='J': band_col=14
        elif band=='H': band_col=15
        elif band=='K': band_col=16
        else:
            raise ValueError(f'band variable {band} is in an incorrect format.')

        data = np.loadtxt(loc, usecols=(1, 2, 3, 4, band_col))
        mTurb= data[:, 0]
        logg = data[:, 1]
        Teff = data[:, 2]
        MH   = data[:, 3]

        if isinstance(MH_range, np.ndarray):
            MH_mask = (MH >= MH_range[0]) & (MH <= MH_range[1])
        else:
            MH_mask = MH == MH_range
            data = np.delete(data, 3, 1)
        if isinstance(Trange, np.ndarray):
            T_mask = (Teff >= Trange[0]) & (Teff <= Trange[1])
        else:
            T_mask = Teff == Trange
            data = np.delete(data, 2, 1)
        if isinstance(logg_range, np.ndarray):
            lg_mask = (logg >= logg_range[0]) & (logg <= logg_range[1])
        else:
            lg_mask = logg == logg_range
            data = np.delete(data, 1, 1)
        if isinstance(mTurb_range, np.ndarray):
            mT_mask = (mTurb >= mTurb_range[0]) & (mTurb <= mTurb_range[1])
        else:
            mT_mask = mTurb == mTurb_range
            data = np.delete(data, 0, 1)  # delete column with singular data value

        mask = lg_mask & T_mask & MH_mask & mT_mask
    else:
        raise IOError("Unknown datafile structure or wrongly defined location.")

    data = data[mask, :]
    return data


def interpolate_linear_limbd(
        logg: float, Teff: float, MH: float, mTurb: float, logg_range=np.array([0, 7]), Trange=np.array([2000, 9000]),
        MH_range=-0.5, mTurb_range=2.0, loc='../Data/tables/atlasco.dat', band='V'
):
    data = read_linear_limbd_param(logg_range, Trange, MH_range, mTurb_range, loc, band)
    vals = data[:, -1]
    points = data[:, 0:-1]
    eval_points = np.array([])
    # # Flexibility depending on size of parameter space (how many have grid points) and table structure used
    if 'atlasco.dat' in loc:
        if isinstance(mTurb_range, np.ndarray):
            eval_points = np.append(eval_points, mTurb)
        if isinstance(logg_range, np.ndarray):
            eval_points = np.append(eval_points, logg)
        if isinstance(Trange, np.ndarray):
            eval_points = np.append(eval_points, Teff)
        if isinstance(MH_range, np.ndarray):
            eval_points = np.append(eval_points, MH)
    else:
        raise IOError("Unknown datafile structure or wrongly defined location.")

    eval_points = np.reshape(eval_points, (1, eval_points.size))
    res = interp.griddata(points, vals, eval_points, method='cubic')

    return res[0]


def estimate_linear_limbd(
        wavelength_range: np.ndarray or tuple, logg: float, T: float, MH: float, mTurb: float,
        logg_range=np.array([0, 7]), Trange=np.array([2000, 9000]), MH_range=-0.5, mTurb_range=2.0,
        loc='../Data/tables/atlasco.dat', print_bool=False
):
    """
    Estimates linear limb-darkening coefficient by interpolating table values by A. Claret 2000
    https://ui.adsabs.harvard.edu/abs/2000A%26A...363.1081C/abstract
    Default table is the atlasco.dat table from https://cdsarc.unistra.fr/viz-bin/cat/J/A+A/363/1081
    Chooses an approximate photometric system by comparing wavelength range with photometric FWHM around effective max.
    """
    photometric_system = estimate_photometric_system(wavelength_range)
    linear_limbd_coef = interpolate_linear_limbd(logg, T, MH, mTurb, logg_range, Trange, MH_range, mTurb_range,
                                                 loc, photometric_system)
    if print_bool:
        print('Approximate Photometric system: ', photometric_system)
        print(linear_limbd_coef)
    return linear_limbd_coef


def main():
    estimate_linear_limbd((5300, 5700), 2.78, 5042, -0.49, 2.0, print_bool=True)
    estimate_linear_limbd((5300, 5700), 4.58, 5621, -0.49, 2.0, print_bool=True)


if __name__ == "__main__":
    main()




import numpy as np
from scipy.signal import fftconvolve
import lmfit
import scipy.constants as scc
from sb2sep.storage_classes import FitParameters


def gaussian_profile(
        velocities: np.ndarray, amplitude: float, radial_velocity_cm: float, gaussian_sigma: float,
        continuum_constant: float, spectrum_broadening_sigma: float
):
    n = velocities.size
    profile_values = np.ones(n) * continuum_constant

    profile_values += amplitude * np.exp(-(velocities-radial_velocity_cm)**2 / (2*gaussian_sigma**2))

    # Broadening by resolution and smoothing factor
    scaled_width = np.sqrt(2*np.pi) * spectrum_broadening_sigma
    broadening_vals = np.exp(-0.5 * (velocities/spectrum_broadening_sigma)**2) / scaled_width

    return fftconvolve(profile_values, broadening_vals, mode='same')


def weight_function(
        velocities: np.ndarray, broadening_function_values: np.ndarray, velocity_fit_half_width: float,
        radial_velocity_guess: float
):
    if radial_velocity_guess is None:
        peak_idx = np.argmax(broadening_function_values)
        mask = (velocities > velocities[peak_idx] - velocity_fit_half_width) & \
               (velocities < velocities[peak_idx] + velocity_fit_half_width)
    else:
        mask = (velocities > radial_velocity_guess - velocity_fit_half_width) & \
               (velocities < radial_velocity_guess + velocity_fit_half_width)

    weight_function_values = np.zeros(broadening_function_values.size)
    weight_function_values[mask] = 1.0
    return weight_function_values


def get_fit_parameter_values(parameters: lmfit.Parameters):
    amplitude = parameters['amplitude'].value
    radial_velocity_cm = parameters['radial_velocity_cm'].value
    gaussian_sigma = parameters['gaussian_sigma'].value
    continuum_constant = parameters['continuum_constant'].value
    spectrum_broadening_sigma = parameters['spectrum_broadening_sigma'].value
    return amplitude, radial_velocity_cm, gaussian_sigma, continuum_constant, spectrum_broadening_sigma


def compare_broadening_function_with_profile(
        parameters: lmfit.Parameters, velocities: np.ndarray, broadening_function_values: np.ndarray,
        weight_function_values: np.ndarray
):
    parameter_vals = get_fit_parameter_values(parameters)
    comparison = broadening_function_values - gaussian_profile(velocities, *parameter_vals)
    return weight_function_values * comparison


def fitting_routine_gaussian_profile(
        velocities: np.ndarray, broadening_function_values: np.ndarray, fitparams: FitParameters,
        smooth_sigma: float, dv: float, print_report=False, compare_func=compare_broadening_function_with_profile
):
    speed_of_light = scc.c / 1000  # in km/s
    spectrum_broadening_sigma = np.sqrt(
        ((speed_of_light/fitparams.spectral_resolution)/(2.354*dv))**2 + (smooth_sigma/dv)**2
    )   # contribution just from spectral resolution and smoothing of broadening function
    params = lmfit.Parameters()

    weight_function_values = weight_function(
        velocities, broadening_function_values, fitparams.velocity_fit_width, fitparams.RV
    )
    peak_idx = np.argmax(broadening_function_values*weight_function_values)
    params.add('amplitude', value=broadening_function_values[peak_idx], min=0.0)
    if fitparams.RV is None:
        params.add(
            'radial_velocity_cm', value=velocities[peak_idx], min=velocities[peak_idx]-fitparams.velocity_fit_width,
            max=velocities[peak_idx]+fitparams.velocity_fit_width
        )
    else:
        params.add(
            'radial_velocity_cm', value=fitparams.RV, min=fitparams.RV-fitparams.velocity_fit_width,
            max=fitparams.RV+fitparams.velocity_fit_width
        )
    if fitparams.vsini_vary_limit is not None:
        params.add(
            'gaussian_sigma', value=fitparams.vsini, vary=fitparams.vary_vsini,
            min=fitparams.vsini - fitparams.vsini*fitparams.vsini_vary_limit,
            max=fitparams.vsini + fitparams.vsini*fitparams.vsini_vary_limit
        )
    else:
        params.add('gaussian_sigma', value=fitparams.vsini, vary=fitparams.vary_vsini)
    params.add(
        'continuum_constant', value=fitparams.continuum, min=np.min(broadening_function_values),
        max=np.max(broadening_function_values), vary=fitparams.vary_continuum
    )
    params.add('spectrum_broadening_sigma', value=spectrum_broadening_sigma, vary=False)

    fit = lmfit.minimize(
        compare_func, params, args=(velocities, broadening_function_values, weight_function_values),
        xtol=1E-8, ftol=1E-8, max_nfev=500
    )

    if print_report:
        print(lmfit.fit_report(fit, show_correl=False))

    parameter_vals = get_fit_parameter_values(fit.params)
    model = gaussian_profile(velocities, *parameter_vals)

    return fit, model
